# 第十二章 图

> 译者：[yuanrw](https://github.com/yuanrw)

# 第十二章 图

在计算机科学中，图是一种代表数学关系的数据结构，它包含一个*顶点（节点）*的集合和一个*边*的集合，每一个顶点和一条边组成了一个向量。如果图中的每条边是无方向的，那就是一个无向图，如果边是有方向的，那就是一个有向图，在有向图中每一条边都从一个顶点出发到另一个顶点。例如我们有顶点v和w，那么连接v和w的边就用(v,w)来表示，如果这是一条无向边，那么也可以用{v,w}表示，如果是有向边，边的方向是从v出发到w，则用[v,w]表示。（v,w）是指连接顶点v和顶点w的一条边，如果(v,w)是无向的，那么我们就说v和w是相邻顶点，顶点的*度*（degree）是指和这个顶点相关联的边的条数。有向图中，*入度*是指进入该顶点的边的条数，出度是指从该顶点出发的边的条数。一般来说，一条边所连接的两个顶点是不一样的，一条边不会从一个顶点出发又回到这个顶点。

如果某一个图的顶点集和边集分别是图G的顶点集的子集和边集的子集的图，那这个图就叫做G的子图。

若存在一个顶点序列![v0](http://latex.codecogs.com/gif.latex?\\$$v_0$$}),![v1](http://latex.codecogs.com/gif.latex?\\$$v_1$$}),…,![vk-1](http://latex.codecogs.com/gif.latex?\\$$v_k-1$$})其中v=![v0](http://latex.codecogs.com/gif.latex?\\$$v_0$$})，![v'](http://latex.codecogs.com/gif.latex?\\$$v'$$})=![vk-1](http://latex.codecogs.com/gif.latex?\\$$v_k-1$$})，序列中任意两个顶点组成的边($$v_i$$,$$v_i+1$$)都在图中，那么这个序列就叫做顶点v到![v'](http://latex.codecogs.com/gif.latex?\\$$v'$$})的一条*路径*，路径的长度（该路径上边的数目）k![ge](http://latex.codecogs.com/gif.latex?\\$$\geq$$})  0，这条定义对有向图和无向图都适用；在有向图中，路径是有方向的。如果路径中没有重复顶点，那我们就称它为*简单路径*。当k > 1且v=![v'](http://latex.codecogs.com/gif.latex?\\$$v'$$})，那么它就是*环*，如果![v0](http://latex.codecogs.com/gif.latex?\\$$v_0$$})…,![vk-2](http://latex.codecogs.com/gif.latex?\\$$v_k-2$$})是不重复的，我们就称它为一个*简单环*；在无向图中，环还必须满足另一个条件：环中不能有两条重复的边。一个没有环的图我们称之为*无环图*。

如果从顶点v到顶点![v'](http://latex.codecogs.com/gif.latex?\\$$v'$$})有路径，那么![v'](http://latex.codecogs.com/gif.latex?\\$$v'$$})和v就是*连通的*。在无向图中，如果存在一个子图，其中任何两个顶点都是联通的，并且子图外的没有顶点与子图内的顶点连通，那么这个子图就是一个*连通分量*。如果一个无向图中只有一个连通分量，那么这个图就是一个*连通图*（即连通分量中包含了左右顶点）。

有向图的连通分量和无向图中一样，唯一的区别是所有的边都是有向边。有向图中如果有一子图，子图中任意两个顶点都是连通的，那么这个子图叫做*强连通分量*。见图12.1和图12.2。

![12.1](https://github.com/yuanrw/cs61b-textbook-zh/blob/ch12/zh/img/12.1.png)

图12.1：一个无向图。星标的边与顶点1，2相连，顶点4的度为0；顶点3，7，8和9的度为1；顶点2和6的度为2；顶点0和度的度为3。被虚线包围的是连通分量；由于图中连通分量的数量大于1，因此这不是连通图。序列[2,1,0,3]顶点2到顶点3的一条路径。路径[2,1,0,2]是一个环。图中唯一包含顶点4的路径是[4]，路径长度为0。最右图的连通分量是一个无环图，也叫*自由树*。

![12.2](https://github.com/yuanrw/cs61b-textbook-zh/blob/ch12/zh/img/12.2.png)

图12.2：一个有向图。被虚线包围的是连通分量。顶点5，6和7形成了一个强连通分量。其他的单个顶点自身也形成了强连通分量。最左图是一个无环图。顶点0和4的入度都为0；顶点1，2和5-8的入度都为1；顶点3的入度为3.顶点3和8的出度为0；顶点1，2，4，5和7的出度为1；顶点0和6的出度为2。

一个无向连通无环图被称为*自由树*（图中任意两个顶点之间只有一条简单路径）。如果一个无向图中任意两个顶点之间有至少两条简单路径，那么这个图是*变连通的*(biconnected)。

在实际运用中，我们用图中的边上记录一些信息。例如，如果顶点表示城市，那么边就表示道路，我们可能希望在边上记录城市间的距离。又例如顶点表示泵站，边表示管道，那我们可能希望在边上记录水的容量。这些数字信息我们叫做*权重*。

## 12.1 一名程序员的规范

对于处理图的程序，现在没有一个单独鲜明的规范，因为不同的算法的不同需求可能会对合适的描述和方便支持这些描述的操作产生巨大的影响。但是处于教学目的，图12.3给出了一个通用的有向图的“一刀切”抽象示例，图12.4对无向图也给出了同样的抽象示例。其思想是顶点和边由自然数（非负整数）表示。而想要添加在顶点或边上的任何附加数据（例如包含更多信息的标签或权重）都可以按照以顶点或边号为索引的附加数组的形式，加在旁边。